"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.useVirtualOverflowYDyn = exports.useCalcVirtualOverflowDyn = exports.virtualOverflowCalcItems = exports.virtualOverflowCalcVisibleRect = void 0;
const react_1 = require("react");
const small_utils_1 = require("./small-utils");
function virtualOverflowCalcVisibleRect(element) {
    const elementRect = element.getBoundingClientRect();
    const visibleRect = {
        top: elementRect.top,
        left: elementRect.left,
        bottom: elementRect.bottom,
        right: elementRect.right
    };
    // clip to window
    visibleRect.top = Math.max(visibleRect.top, 0);
    visibleRect.left = Math.max(visibleRect.left, 0);
    visibleRect.bottom = Math.min(visibleRect.bottom, window.innerHeight);
    visibleRect.right = Math.min(visibleRect.right, window.innerWidth);
    let currentElement = element.parentElement;
    while (currentElement) {
        const rect = currentElement.getBoundingClientRect();
        // clip to the parent's rectangle
        visibleRect.top = Math.max(visibleRect.top, rect.top);
        visibleRect.left = Math.max(visibleRect.left, rect.left);
        visibleRect.bottom = Math.min(visibleRect.bottom, rect.bottom);
        visibleRect.right = Math.min(visibleRect.right, rect.right);
        currentElement = currentElement.parentElement;
    }
    const contentOffsetTop = visibleRect.top - elementRect.top;
    const contentOffsetLeft = visibleRect.left - elementRect.left;
    return {
        top: visibleRect.top,
        left: visibleRect.left,
        bottom: visibleRect.bottom,
        right: visibleRect.right,
        contentOffsetTop,
        contentOffsetLeft,
        contentVisibleHeight: visibleRect.bottom - visibleRect.top,
        contentVisibleWidth: visibleRect.right - visibleRect.left,
    };
}
exports.virtualOverflowCalcVisibleRect = virtualOverflowCalcVisibleRect;
// contentOffsetStart = visibleRect.contentOffsetTop
// contentVisibleSize = visibleRect.contentVisibleHeight
function virtualOverflowCalcItems(contentOffsetStart, contentVisibleSize, itemSizes, overscanItemsCount, itemsLength) {
    let itemStart = 0;
    for (; itemStart < itemsLength && (!itemSizes[itemStart] || (contentOffsetStart < itemSizes[itemStart]?.offset)); ++itemStart) {
    }
    const endOffset = contentOffsetStart + contentVisibleSize;
    let itemEnd = itemStart;
    for (; itemEnd < itemsLength && (!itemSizes[itemEnd] || (endOffset < itemSizes[itemEnd]?.offset)); ++itemEnd) {
    }
    let itemLen = itemEnd - itemStart;
    itemStart = Math.max(0, itemStart - overscanItemsCount);
    itemLen = Math.max(0, Math.min(itemsLength - itemStart, itemLen + overscanItemsCount + overscanItemsCount));
    return { itemStart, itemLen };
}
exports.virtualOverflowCalcItems = virtualOverflowCalcItems;
function useCalcVirtualOverflowDyn(params, deps) {
    const { containerRef, itemsLengthX, itemsLengthY, sizeCacheRef, overscanItemsCount = 3, calcVisibleRect = virtualOverflowCalcVisibleRect } = params;
    const [itemSlice, setItemSlice] = (0, react_1.useState)({
        topStartIndex: 0,
        lengthY: 0,
        leftStartIndex: 0,
        lengthX: 0,
    });
    const { requestFrame: updateViewRect, cancelFrame } = (0, react_1.useMemo)(() => (0, small_utils_1.rvoDebounceAnimationFrame)((frameTime) => {
        if (!containerRef.current)
            return;
        const { children } = params.containerRef.current;
        for (let i = 0; i < children.length; ++i) {
            const el = children.item(i);
            if (!el)
                continue;
            const reactElIndex = getReactFiberIndex(el);
            const elRect = el.getBoundingClientRect();
            const sizeIndex = itemSlice.topStartIndex + reactElIndex;
            const prevSize = params.sizeCacheRef.current[sizeIndex - 1] || { self: 0, offset: 0 };
            params.sizeCacheRef.current[sizeIndex] = {
                offset: prevSize.offset + prevSize.self,
                self: elRect.height,
            };
        }
        params.sizeCacheRef.current[children.length] = params.sizeCacheRef.current[children.length - 1];
        const visibleRect = calcVisibleRect(containerRef.current, frameTime);
        const newItemSlice = {
            topStartIndex: 0,
            lengthY: 1,
            leftStartIndex: 0,
            lengthX: 1,
        };
        if (itemsLengthY !== undefined) {
            const verticalSlice = virtualOverflowCalcItems(visibleRect.contentOffsetTop, visibleRect.contentVisibleHeight, sizeCacheRef.current, overscanItemsCount, itemsLengthY);
            newItemSlice.topStartIndex = verticalSlice.itemStart;
            newItemSlice.lengthY = verticalSlice.itemLen;
        }
        if (itemsLengthX !== undefined) {
            const horizontalSlice = virtualOverflowCalcItems(visibleRect.contentOffsetLeft, visibleRect.contentVisibleWidth, sizeCacheRef.current, overscanItemsCount, itemsLengthX);
            newItemSlice.leftStartIndex = horizontalSlice.itemStart;
            newItemSlice.lengthX = horizontalSlice.itemLen;
        }
        setItemSlice(newItemSlice);
    }), [containerRef.current, itemsLengthX, itemsLengthY, sizeCacheRef, ...deps]);
    (0, react_1.useLayoutEffect)(() => {
        window.addEventListener('scroll', updateViewRect, { capture: true, passive: true });
        window.addEventListener('resize', updateViewRect, { capture: true, passive: true });
        window.addEventListener('orientationchange', updateViewRect, { capture: true, passive: true });
        updateViewRect();
        return () => {
            cancelFrame();
            window.removeEventListener('scroll', updateViewRect);
            window.removeEventListener('resize', updateViewRect);
            window.removeEventListener('orientationchange', updateViewRect);
        };
    }, [containerRef.current, itemsLengthX, itemsLengthY, sizeCacheRef, ...deps]);
    return { itemSlice, updateViewRect };
}
exports.useCalcVirtualOverflowDyn = useCalcVirtualOverflowDyn;
function utilRenderItems1D(itemStart, itemsLength, sizeCacheRef, renderItem) {
    const itemEnd = itemStart + itemsLength;
    const renderedItems = Array.from({ length: itemsLength });
    for (let i = itemStart; i < itemEnd; ++i) {
        renderedItems[i - itemStart] = renderItem(i, sizeCacheRef.current[i]?.offset);
    }
    return renderedItems;
}
let reactFiberField = undefined;
function getReactFiberIndex(el) {
    if (!reactFiberField) {
        const foundKey = Object.keys(el).find(x => x.startsWith('__reactFiber'));
        if (!foundKey) {
            console.error('__reactFiber key not found');
            return 0;
        }
        reactFiberField = foundKey;
    }
    return el[reactFiberField].index;
}
function useVirtualOverflowYDyn(params, deps = []) {
    const { itemSlice, updateViewRect } = useCalcVirtualOverflowDyn(params, deps);
    return {
        renderedItems: utilRenderItems1D(itemSlice.topStartIndex, itemSlice.lengthY, params.sizeCacheRef, params.renderItem),
        updateViewRect,
    };
}
exports.useVirtualOverflowYDyn = useVirtualOverflowYDyn;
//# sourceMappingURL=dynamic.js.map